"""Constants used throughout Farcy."""
import os
import re

__version__ = '1.2'

VERSION_STR = 'farcy v{0}'.format(__version__)

CONFIG_DIR = os.path.expanduser('~/.config/farcy')

MD_VERSION_STR = ('[{0}](https://github.com/appfolio/farcy)'
                  .format(VERSION_STR))

FARCY_COMMENT_START = '_{0}_'.format(MD_VERSION_STR)

NUMBER_RE = re.compile('(\d+)')

APPROVAL_PHRASES = [x.strip() for x in """
Amazing
Bravo
Excellent
Great job
Lookin' good
Outstanding work
Perfect
Spectacular
Tremendous
Well done
Wicked awesome
Winning
Wonderful
Wow
You are awesome
You do not miss a thing
""".split('\n') if x.strip()]

STATUS_CONTEXT = 'farcy'



"""Farcy objects test file."""

from __future__ import print_function
from farcy import objects
from mock import patch
import unittest
import farcy.exceptions as exceptions
from .helper import Struct


class ConfigTest(unittest.TestCase):
    """Tests Config helper."""

    @patch('farcy.objects.ConfigParser')
    @patch('os.path.isfile')
    def _config_instance(self, callback, mock_is_file, mock_config, repo=None,
                         post_callback=None, **overrides):
        mock_is_file.called_with(objects.Config.PATH).return_value = True

        if callback:
            callback(mock_config.return_value)
        config = objects.Config(repo, **overrides)
        if post_callback:
            post_callback(mock_config.return_value)
        self.assertTrue(mock_config.called)
        return config

    def test_cant_change_log_level_if_debug(self):
        config = self._config_instance(None, repo='a/b')
        self.assertNotEqual('DEBUG', config.log_level)
        config.debug = True
        self.assertEqual('DEBUG', config.log_level)
        config.log_level = 'WARNING'
        self.assertEqual('DEBUG', config.log_level)

    def test_config_file_is_overridable(self):
        def callback(mock_config):
            mock_config.items.return_value = {'start_event': '1337'}
        config = self._config_instance(callback, repo='a/b')
        self.assertEqual(1337, config.start_event)
        config.start_event = 10
        self.assertEqual(10, config.start_event)

    def test_config_file_repo_specific_works(self):
        def callback(mock_config):
            mock_config.has_section.return_value = True
            mock_config.items.return_value = {'start_event': '1337'}

        def post_callback(mock_config):
            mock_config.items.assert_called_with('a/b')
        config = self._config_instance(callback, repo='a/b',
                                       post_callback=post_callback)
        self.assertEqual('a/b', config.repository)
        self.assertEqual(1337, config.start_event)

    def test_config_file_values(self):
        def callback(mock_config):
            mock_config.has_section.return_value = False
            mock_config.items.return_value = {
                'start_event': '10', 'debug': True,
                'exclude_paths': 'node_modules,vendor',
                'limit_users': 'balloob,bboe', 'log_level': 'DEBUG',
                'pr_issue_report_limit': '100'}

        def post_callback(mock_config):
            mock_config.items.assert_called_with('DEFAULT')
        config = self._config_instance(callback, repo='a/b',
                                       post_callback=post_callback)
        self.assertEqual('a/b', config.repository)
        self.assertEqual(10, config.start_event)
        self.assertEqual(True, config.debug)
        self.assertEqual({'node_modules', 'vendor'}, config.exclude_paths)
        self.assertEqual({'balloob', 'bboe'}, config.limit_users)
        self.assertEqual('DEBUG', config.log_level)
        self.assertEqual(100, config.pr_issue_report_limit)

    def test_config__overrides(self):
        config = self._config_instance(None, repo='a/b', start_event=1337,
                                       limit_users='bboe')
        self.assertEqual('a/b', config.repository)
        self.assertEqual(1337, config.start_event)
        self.assertEqual({'bboe'}, config.limit_users)

    def test_config__repr(self):
        config = self._config_instance(None, repo='a/b')
        repr_str = ("Config('a/b', comment_group_threshold=3, debug=False, "
                    "exclude_paths=None, exclude_users=None, "
                    "limit_users=None, log_level='ERROR', "
                    "pr_issue_report_limit=128, pull_requests=None, "
                    "start_event=None)")
        self.assertEqual(repr_str, repr(config))

    def test_default_repo_from_config(self):
        def callback(mock_config):
            mock_config.get.return_value = 'appfolio/farcy'
        config = self._config_instance(callback)
        self.assertEqual('appfolio/farcy', config.repository)

    def test_default_repo_from_config_raise_on_invalid(self):
        def callback(mock_config):
            mock_config.get.return_value = 'invalid_repo'
        with self.assertRaises(exceptions.FarcyException):
            self._config_instance(callback)

    def test_raise_if_invalid_log_level(self):
        config = self._config_instance(None, repo='a/b')
        with self.assertRaises(exceptions.FarcyException):
            config.log_level = 'invalid_log_level'

    def test_raise_if_invalid_repository(self):
        config = self._config_instance(None, repo='a/b')
        with self.assertRaises(exceptions.FarcyException):
            config.repository = 'invalid_repo'

    def test_raise_if_setting_blacklist_with_whitelist_already_set(self):
        config = self._config_instance(None, limit_users=['a'], repo='a/b')
        with self.assertRaises(exceptions.FarcyException):
            config.exclude_users = ['b']

    def test_raise_if_setting_whitelist_with_blacklist_already_set(self):
        config = self._config_instance(None, exclude_users=['a'], repo='a/b')
        with self.assertRaises(exceptions.FarcyException):
            config.limit_users = ['b']

    def test_setting_repo(self):
        config = self._config_instance(None, repo='a/b')
        self.assertEqual('a/b', config.repository)
        config.repository = 'appfolio/farcy'
        self.assertEqual('appfolio/farcy', config.repository)

    def test_setting_values_via_dict(self):
        config = self._config_instance(None, repo='appfolio/farcy')
        data = {
            'start_event': 1000,
            'debug': False,
            'exclude_paths': {'npm_modules', 'vendor'},
            'limit_users': {'balloob', 'bboe'},
            'log_level': 'WARNING',
            'pr_issue_report_limit': 100
        }

        config.override(**data)
        for attr, value in data.items():
            self.assertEqual(value, getattr(config, attr))

    def test_user_allowed_with_no_lists_set(self):
        config = self._config_instance(None, repo='a/b')
        self.assertTrue(config.user_allowed('balloob'))

    def test_user_allowed_when_not_blacklisted(self):
        config = self._config_instance(None, exclude_users=['a'], repo='a/b')
        self.assertTrue(config.user_allowed('balloob'))

    def test_user_allowed_when_whitelisted(self):
        config = self._config_instance(None, repo='a/b')
        config.limit_users = ['bboe', 'balloob']
        self.assertTrue(config.user_allowed('balloob'))
        self.assertFalse(config.user_allowed('appfolio'))

    def test_user_not_allowed_when_blacklisted(self):
        config = self._config_instance(None, exclude_users=['balloob'],
                                       repo='a/b')
        self.assertFalse(config.user_allowed('balloob'))

    def test_user_not_allowed_when_not_whitelisted(self):
        config = self._config_instance(None, limit_users=['a'],
                                       repo='a/b')
        self.assertFalse(config.user_allowed('balloob'))


class ErrorMessageTest(unittest.TestCase):
    def setUp(self):
        self.message = objects.ErrorMessage('Dummy Message', 2)

    def add_lines(self, on_github, *lines):
        for line in lines:
            self.message.track(line, on_github)

    def test_messages__group_consequtive(self):
        self.add_lines(False, 1, 2, 3)
        self.assertEqual([(1, 'Dummy Message <sub>3x spanning 3 lines</sub>')],
                         list(self.message.messages()))

    def test_messages__group_span(self):
        self.add_lines(False, 1, 3, 5)
        self.assertEqual([(1, 'Dummy Message <sub>3x spanning 5 lines</sub>')],
                         list(self.message.messages()))

    def test_messages__group_span__existing_group(self):
        self.add_lines(False, 1, 3, 5)
        self.message.track_group(1, 3)
        self.assertEqual([], list(self.message.messages()))

    def test_messages__no_grouping(self):
        self.add_lines(False, 1, 4, 100, 105)
        self.add_lines(True, 2, 4, 101, 106)
        self.assertEqual([(1, 'Dummy Message'), (100, 'Dummy Message'),
                          (105, 'Dummy Message')],
                         list(self.message.messages()))

    def test_messages__no_messages(self):
        self.assertEqual([], list(self.message.messages()))

    def test_track__return_value(self):
        self.assertEqual(self.message, self.message.track(16))

    def test_track_group__return_value(self):
        self.assertEqual(self.message, self.message.track_group(16, 2))


class ErrorTrackerTest(unittest.TestCase):
    def setUp(self):
        self.tracker = objects.ErrorTracker([], 2)

    def test_initial_values(self):
        self.assertEqual(0, self.tracker.github_message_count)
        self.assertEqual(0, self.tracker.new_issue_count)

    def test_no_issues(self):
        self.assertEqual([], list(self.tracker.errors('DummyFile')))

        comment = Struct(body='Regular comment', path='DummyFile', position=16)
        self.tracker.from_github_comments([comment])

        self.assertEqual(0, self.tracker.github_message_count)
        self.assertEqual(0, self.tracker.hidden_issue_count)
        self.assertEqual(0, self.tracker.new_issue_count)
        self.assertEqual([], list(self.tracker.errors('DummyFile')))

    def test_grouped_issue(self):
        comment = Struct(body=('_[farcy \n* MatchingError <sub>3x spanning 4 '
                               'lines</sub>'), path='DummyFile', position=16)
        self.tracker.from_github_comments([comment])
        self.tracker.track('MatchingError', 'DummyFile', 16)
        self.tracker.track('MatchingError', 'DummyFile', 18)
        self.tracker.track('MatchingError', 'DummyFile', 19)

        self.assertEqual(1, self.tracker.github_message_count)
        self.assertEqual(0, self.tracker.hidden_issue_count)
        self.assertEqual(3, self.tracker.new_issue_count)
        self.assertEqual([], list(self.tracker.errors('DummyFile')))

    def test_one_issue_with_duplicate(self):
        self.assertEqual([], list(self.tracker.errors('DummyFile')))

        comment = Struct(body='_[farcy \n* MatchingError', path='DummyFile',
                         position=16)
        self.tracker.from_github_comments([comment])

        self.assertEqual(1, self.tracker.github_message_count)
        self.assertEqual(0, self.tracker.hidden_issue_count)
        self.assertEqual(0, self.tracker.new_issue_count)
        self.assertEqual([], list(self.tracker.errors('DummyFile')))

        self.tracker.track('MatchingError', 'DummyFile', 16)
        self.assertEqual(1, self.tracker.github_message_count)
        self.assertEqual(0, self.tracker.hidden_issue_count)
        self.assertEqual(1, self.tracker.new_issue_count)
        self.assertEqual([], list(self.tracker.errors('DummyFile')))

    def test_one_unique_issue__different_line(self):
        comment = Struct(body='_[farcy \n* MatchingError', path='DummyFile',
                         position=16)
        self.tracker.from_github_comments([comment])
        self.tracker.track('MatchingError', 'DummyFile', 16)
        self.tracker.track('Non MatchingError', 'DummyFile', 17)

        self.assertEqual(1, self.tracker.github_message_count)
        self.assertEqual(2, self.tracker.new_issue_count)
        self.assertEqual([(17, ['Non MatchingError'])],
                         list(self.tracker.errors('DummyFile')))

    def test_one_unique_issue__same_line(self):
        comment = Struct(body='_[farcy \n* MatchingError', path='DummyFile',
                         position=16)
        self.tracker.from_github_comments([comment])
        self.tracker.track('MatchingError', 'DummyFile', 16)
        self.tracker.track('Non MatchingError', 'DummyFile', 16)

        self.assertEqual(1, self.tracker.github_message_count)
        self.assertEqual(2, self.tracker.new_issue_count)
        self.assertEqual([(16, ['Non MatchingError'])],
                         list(self.tracker.errors('DummyFile')))

    def test_only_hidden_issues(self):
        comment = Struct(body='_[farcy \n* MatchingError', path='DummyFile',
                         position=0)
        self.tracker.from_github_comments([comment])

        self.assertEqual(0, self.tracker.github_message_count)
        self.assertEqual(1, self.tracker.hidden_issue_count)
        self.assertEqual(0, self.tracker.new_issue_count)
        self.assertEqual([], list(self.tracker.errors('DummyFile')))



"""Farcy handlers test file."""

from __future__ import print_function
import os
import unittest
from farcy.exceptions import HandlerException
import farcy.handlers


class ExtHandlerTest(unittest.TestCase):

    """Tests common Farcy handler extension methods."""

    def setUp(self):
        self.cls = farcy.handlers.ExtHandler
        self.cls.BINARY_VERSION = '1.0'
        self.cls.BINARY = 'FOOBAR'

    def test_constructor__binary_not_installed(self):
        CLS2 = type('CLS2', (farcy.handlers.ExtHandler,), {})
        with self.assertRaises(farcy.exceptions.HandlerException) as cm:
            self.instance = CLS2()
        self.assertEqual('FOOBAR is not installed.',
                         str(cm.exception))

    def test_constructor__binary_permission_denied(self):
        CLS2 = type('CLS2', (farcy.handlers.ExtHandler,), {})
        CLS2.BINARY = '/tmp'
        with self.assertRaises(farcy.exceptions.HandlerException) as cm:
            self.instance = CLS2()
        self.assertEqual('/tmp cannot be executed.',
                         str(cm.exception))

    def test_constructor__binary_required(self):
        CLS2 = type('CLS2', (farcy.handlers.ExtHandler,), {})
        CLS2.BINARY = None
        with self.assertRaises(farcy.exceptions.HandlerException) as cm:
            self.instance = CLS2()
        self.assertEqual('CLS2 does not have a binary specified.',
                         str(cm.exception))

    def test_constructor__subclass_required(self):
        with self.assertRaises(farcy.exceptions.HandlerException) as cm:
            self.instance = self.cls()
        self.assertEqual('Base class `ExtHandler` must be extended.',
                         str(cm.exception))

    def test_verify_version(self):
        self.assertEqual(None, self.cls.verify_version('1.0'))
        self.assertEqual(None, self.cls.verify_version('1.0', True))
        self.assertEqual(None, self.cls.verify_version('1.0.0'))
        self.assertEqual(None, self.cls.verify_version('1.0.1'))
        self.assertEqual(None, self.cls.verify_version('1.1.1'))
        self.assertEqual(None, self.cls.verify_version('2'))

    def test_verify_version__expected_too_small(self):
        with self.assertRaises(farcy.exceptions.HandlerException) as cm:
            self.cls.verify_version('0.9')
        self.assertEqual('Expected FOOBAR >= 1.0, found 0.9',
                         str(cm.exception))

        with self.assertRaises(farcy.exceptions.HandlerException) as cm:
            self.cls.verify_version('1.0 ')  # Contains space
        self.assertEqual('Expected FOOBAR >= 1.0, found 1.0 ',
                         str(cm.exception))

    def test_verify_version__not_exact_match(self):
        with self.assertRaises(farcy.exceptions.HandlerException) as cm:
            self.cls.verify_version('1.0.1', True)
        self.assertEqual('Expected FOOBAR 1.0, found 1.0.1',
                         str(cm.exception))


class FarcyTest(unittest.TestCase):

    """Provides helpers for various FarcyTest classes."""

    def shortDescription(self):
        """Disable the short description (docstring) output."""
        return None

    def path(self, filename):
        """Return a path to the example file."""
        return os.path.join(os.path.dirname(__file__), 'examples', filename)

    def config(self, linter):
        """Return config path for linter."""
        return os.path.join(
            os.path.dirname(__file__), 'configs',
            'handler_{0}.conf'.format(linter.__class__.__name__.lower()))


class ESLintTest(FarcyTest):

    """Tests for the ESLint Handler."""

    def setUp(self):
        """Set up helpers used for each test case."""
        self.linter = farcy.handlers.ESLint()
        self.linter.config_file_path = self.config(self.linter)

    def test_perfect_file(self):
        """There should be no issues."""
        errors = self.linter.process(self.path('no_issue.js'))
        self.assertEqual({}, errors)

    def test_single_error(self):
        """A single error should be returned."""
        errors = self.linter.process(self.path('single_issue.js'))
        self.assertEqual({3: ['Unexpected console statement. (no-console)']},
                         errors)

    def test_invalid_syntax(self):
        """Test an error is returned for correct line when syntax error."""
        errors = self.linter.process(self.path('invalid_syntax.js'))
        self.assertEqual({3: ['Parsing error: Unexpected token name']},
                         errors)


class Flake8Test(FarcyTest):

    """Tests for the Flake8 Handler."""

    def setUp(self):
        """Set up helpers used for each test case."""
        self.process = farcy.handlers.Flake8().process

    def test_perfect_file(self):
        """There should be no issues."""
        errors = self.process(self.path('no_issue.py'))
        self.assertEqual({}, errors)

    def test_single_error(self):
        """A single error should be returned."""
        errors = self.process(self.path('single_issue.py'))
        self.assertEqual({3: ['1: E302 expected 2 blank lines, found 1']},
                         errors)


class Pep257Test(FarcyTest):

    """Tests for the Pep257 Handler."""

    def setUp(self):
        """Set up helpers used for each test case."""
        self.process = farcy.handlers.Pep257().process

    def test_perfect_file(self):
        """There should be no issues."""
        errors = self.process(self.path('no_issue.py'))
        self.assertEqual({}, errors)

    def test_single_error(self):
        """A single error should be returned."""
        errors = self.process(self.path('single_issue.py'))
        self.assertEqual({3: [('D211: No blank lines allowed before class '
                               'docstring (found 1)')]},
                         errors)


class RubocopTest(FarcyTest):

    """Tests for the Rubocop Handler."""

    def setUp(self):
        """Set up helpers used for each test case."""
        self.linter = farcy.handlers.Rubocop()
        self.linter.config_file_path = self.config(self.linter)

    def test_perfect_file(self):
        """There should be no issues."""
        errors = self.linter.process(self.path('no_issue.rb'))
        self.assertEqual({}, errors)

    def test_single_error(self):
        """A single error should be returned.

           Uses config to disable 1 issue.
        """
        errors = self.linter.process(self.path('single_issue.rb'))
        self.assertEqual({3: [('Style/DefWithParentheses: Omit the parentheses'
                               ' in defs when the method doesn\'t accept any '
                               'arguments.')]},
                         errors)


class SCSSLintTest(FarcyTest):
    """Tests for the SCSSLint Handler."""

    def setUp(self):
        """Set up helpers used for each test case."""
        self.linter = farcy.handlers.SCSSLint()
        self.linter.config_file_path = self.config(self.linter)

    def test_perfect_file(self):
        """There should be no issues."""
        errors = self.linter.process(self.path('no_issue.scss'))
        self.assertEqual({}, errors)

    def test_single_error(self):
        """A single error should be returned.

           Uses config to disable 1 issue.
        """
        errors = self.linter.process(self.path('single_issue.scss'))
        self.assertEqual({1: [('SelectorFormat: Selector `test_class` should '
                               'be written in lowercase with hyphens')]},
                         errors)

    def test_linting_exception(self):
        """Test an error is raised with useful information if linting fails."""
        try:
            self.linter.process(self.path('linting_exception.scss'))
        except HandlerException as exc:
            self.assertEqual('Error occurred during linting: Syntax Error: '
                             'Invalid CSS after ".broken-class '
                             '{": expected "}", was "" (line 2, column 1)',
                             str(exc))



"""Defines the exceptions used in the farcy package."""


class FarcyException(Exception):
    """Farcy root exception class."""

    def __str__(self):
        """Return the exception's class name."""
        retval = super(FarcyException, self).__str__()
        return retval or self.__class__.__name__


class HandlerException(FarcyException):
    """Farcy handler primary exception."""


class HandlerNotReady(HandlerException):
    """Exception indicating that a handler is not ready for use."""



"""Defines handlers for various file types."""

from __future__ import print_function
from collections import defaultdict
from subprocess import CalledProcessError, STDOUT, check_output
from update_checker import parse_version
import json
import logging
import os
import re
from .const import CONFIG_DIR
from .exceptions import HandlerException, HandlerNotReady


# src: http://stackoverflow.com/a/11270665/176978
try:
    from subprocess import DEVNULL
except ImportError:
    DEVNULL = open(os.devnull, 'wb')


class ExtHandler(object):
    """An abstract class that provides the file handler interface.

    Subclasses must define the extension(s) that the class provides feedback
    upon via the ``EXTENSIONS`` class method.

    ``BINARY`` is the name of an executable binary to look for.
    ``BINARY_VERSION`` is version of the binary expected.

    """

    BINARY = None
    BINARY_VERSION = None
    EXTENSIONS = []
    OUTPUT = 'stdout'

    @staticmethod
    def execute(args, stderr=DEVNULL):
        """Return output of argument execution ignoring status code."""
        try:
            return check_output(args, stderr=stderr).decode('utf-8')
        except CalledProcessError as exc:
            return exc.output.decode('utf-8')

    @classmethod
    def verify_version(cls, installed, exact=False):
        """Raise HandlerException if the installed version does not match.

        :param installed: The installed version number.
        :param exact: Raise HandlerException when there is not an exact
            match. Note that 0.27.0 is considered exact to 0.27.

        """
        exp = parse_version(cls.BINARY_VERSION)
        inp = parse_version(installed)
        op = None
        if exact and exp != inp:
            op = ''
        elif exp > inp:
            op = '>= '
        if op is not None:
            raise HandlerException(
                'Expected {0} {1}{2}, found {3}'
                .format(cls.BINARY, op, cls.BINARY_VERSION, installed))

    def __init__(self, on_demand=False):
        """A handler's constructor is called only once upon farcy start-up.

        :param on_demand: When true, plugins that are not usable on start-up
            are still loaded, and will be tested for use on-demand.

        By default this method only calls the ``assert_usable`` instance method
        to see if the plugin's dependencies are available.

        """
        self._logger = logging.getLogger(__name__)
        self.name = type(self).__name__
        try:
            self.assert_usable()
            self._plugin_ready = True
        except HandlerNotReady as exc:
            if on_demand:
                self._logger.warning('{0} is not ready: {1}'
                                     .format(self.name, str(exc)))
            else:
                raise
            self._plugin_ready = False
        path = os.path.join(
            CONFIG_DIR, 'handler_{0}.conf'.format(self.name.lower()))
        self.config_file_path = path if os.path.isfile(path) else None

    def _regex_parse(self, binary_args, stderr=None):
        """Use the sublcasses RE value to parse the returned data."""
        retval = defaultdict(list)
        for (lineno, msg) in self.RE.findall(self.execute(
                [self.BINARY] + binary_args, stderr=stderr)):
            retval[int(lineno)].append(msg)
        return retval

    def assert_usable(self):
        """Raise HandlerException if the handler is not ready for use."""
        if self.name == 'ExtHandler':
            raise HandlerException('Base class `ExtHandler` must be extended.')
        if not self.BINARY:
            raise HandlerException('{0} does not have a binary specified.'
                                   .format(self.name))
        try:
            version = (check_output([self.BINARY, '--version'], stderr=STDOUT)
                       .decode('utf-8'))
        except OSError as exc:
            if exc.errno == 2:
                raise HandlerNotReady('{0} is not installed.'
                                      .format(self.BINARY))
            elif exc.errno == 13:
                raise HandlerException('{0} cannot be executed.'
                                       .format(self.BINARY))
            raise  # Unexpected and unhandled exception
        self.verify_version(self.version_callback(version))

    def process(self, filename):
        """Return a dictionary mapping line numbers to errors.

        The value for each line number in the dictionary should be a list where
        each item of the list is a string containing an error that occurred on
        the line.

        :param filename: The filename to analyze.

        """
        # This method should not be implemented by a subclass. Use _process
        # instead.
        if not self._plugin_ready:
            try:
                self.assert_usable()
                self._plugin_ready = True
            except HandlerNotReady as exc:
                self._logger.warning('{0} is not ready: {1}'
                                     .format(self.name, exc.message))
                return {}
        return self._process(filename)

    def version_callback(self, version):
        """Return a parsed version string for the binary version."""
        return version.strip()


class ESLint(ExtHandler):
    """Provides feedback for JavaScript files using ESLint."""

    BINARY = 'eslint'
    BINARY_VERSION = '1.1.0'
    EXTENSIONS = ['.js', '.jsx']

    def _process(self, filename):
        command = [self.BINARY, '--format', 'json']
        config_path = self.config_file_path
        if config_path:
            command += ['--config', config_path]

        data = json.loads(self.execute(command + [filename]))[0]
        retval = defaultdict(list)

        for offense in data['messages']:
            message = offense['message']
            if offense.get('ruleId'):
                message += ' ({})'.format(offense['ruleId'])
            retval[offense['line']].append(message)
        return retval

    def version_callback(self, version):
        """Remove the 'v' prefix and trailing space."""
        return version[1:].strip()


class Flake8(ExtHandler):
    """Provides feedback for python files using flake8."""

    BINARY = 'flake8'
    BINARY_VERSION = '2.4.1'
    EXTENSIONS = ['.py']
    RE = re.compile('[^:]+:(\d+):([^\n]+)\n')

    def _process(self, filename):
        config_path = self.config_file_path
        command = ['--config', config_path] if config_path else []
        return self._regex_parse(command + [filename])

    def version_callback(self, version):
        """Remove the extra version information."""
        return version.split(' ', 1)[0]


class JSXHint(ExtHandler):
    """Provides feedback for JS/JSX files using jsxhint."""

    BINARY = 'jsxhint'
    BINARY_VERSION = '0.15.0'
    EXTENSIONS = ['.jsx', '.js']
    RE = re.compile('.*:(\d+):\d+: (.*)\n')

    def _process(self, filename):
        command = ['--reporter', 'unix']
        config_path = self.config_file_path
        if config_path:
            command += ['--config', config_path]
        return self._regex_parse(command + [filename])

    def version_callback(self, version):
        """Return a parsed version string for the binary version."""
        return version.split(' ')[1][1:] if ' ' in version else ''


class Pep257(ExtHandler):
    """Provides feedback for python files using pep257."""

    BINARY = 'pep257'
    BINARY_VERSION = '0.5.0'
    EXTENSIONS = ['.py']
    RE = re.compile('[^:]+:(\d+)[^\n]+\n\s+([^\n]+)\n')

    def _process(self, filename):
        return self._regex_parse([filename], stderr=STDOUT)


class Rubocop(ExtHandler):
    """Provides feedback for ruby files using rubocop."""

    BINARY = 'rubocop'
    BINARY_VERSION = '0.50'
    EXTENSIONS = ['.rb']

    def _process(self, filename):
        command = [self.BINARY, '-f', 'j']
        config_path = self.config_file_path
        if config_path:
            command += ['-c', config_path]

        data = json.loads(self.execute(command + [filename]))
        retval = defaultdict(list)
        for offense in data.get('files', [{}])[0].get('offenses', []):
            retval[offense['location']['line']].append(offense['message'])
        return retval


class SCSSLint(ExtHandler):
    """Provides feedback for css and scss files using scss-lint."""

    BINARY = 'scss-lint'
    BINARY_VERSION = '0.43.2'
    EXTENSIONS = ['.css', '.scss']

    def _process(self, filename):
        command = [self.BINARY, '-f', 'JSON']
        config_path = self.config_file_path
        if config_path:
            command += ['-c', config_path]

        data = json.loads(self.execute(command + [filename]))

        retval = defaultdict(list)
        if not data.values():
            return retval
        for offense in next(iter(data.values())):
            if 'linter' not in offense:
                exception_message = (
                    "Error occurred during linting: {reason} "
                    "(line {line}, column {column})"
                ).format(**offense)
                raise HandlerException(exception_message)
            retval[offense['line']].append(
                '{linter}: {reason}'.format(**offense)
            )

        return retval

    def version_callback(self, version):
        """
        Return a parsed version string for the binary version.

        This returns just the semantic versioned portion of the version string.
        """
        return version.split()[1]



"""Farcy class test file."""

from __future__ import print_function
from collections import namedtuple
from datetime import datetime
from farcy import (Config, FARCY_COMMENT_START, Farcy, FarcyException, UTC,
                   main, no_handler_debug_factory)
from mock import MagicMock, call, patch
from requests import ConnectionError
import farcy as farcy_module
import logging
import unittest
from .helper import Struct

Config.PATH = '/dev/null'  # Don't allow the system config file to load.
farcy_module.APPROVAL_PHRASES = ['Dummy Approval']  # Provide only one option.

PFILE_ATTRS = ['contents', 'filename', 'patch', 'status']

MockInfo = namedtuple('Info', ['decoded'])
MockPFile = namedtuple('PFile', PFILE_ATTRS)


def assert_calls(method, *calls):
    method.assert_has_calls(list(calls))
    assert method.call_count == len(calls), "{0} != {1}".format(
        list(calls), method.mock_calls)


def assert_status(farcy, failures=0):
    if failures:
        call2 = call('dummy', 'failure', context='farcy',
                     description='found {0} issue{1}'.format(
                         failures, 's' if failures > 1 else ''))
    else:
        call2 = call('dummy', 'success', context='farcy',
                     description='approves! Dummy Approval!')
    assert_calls(farcy.repo.create_status,
                 call('dummy', 'pending', context='farcy',
                      description='started investigation'), call2)


def mockpfile(**kwargs):
    for attr in PFILE_ATTRS:
        kwargs.setdefault(attr, None)
    return MockPFile(**kwargs)


class FarcyBaseTest(unittest.TestCase):
    def setUp(self):
        logging.disable(logging.CRITICAL)
        self.logger = logging.getLogger('farcy')

    @patch('farcy.UpdateChecker')
    @patch('farcy.objects.get_session')
    def _farcy_instance(self, mock_get_session, mock_update_checker,
                        config=None):
        if config is None:
            config = Config(None)
        if config.repository is None:
            config.repository = 'dummy/dummy'
        farcy = Farcy(config)
        self.assertTrue(mock_get_session.called)
        self.assertTrue(mock_update_checker.called)
        return farcy


class FarcyTest(FarcyBaseTest):
    @patch('farcy.added_lines')
    def test_compute_pfile_stats__added(self, mock_added_lines):
        mock_added_lines.return_value = {13: 10, 15: 20, 18: 100}
        stats = {'added_files': 10, 'added_lines': 10}
        actual = self._farcy_instance()._compute_pfile_stats(
            mockpfile(patch='', status='added'), stats)
        self.assertTrue(mock_added_lines.called)
        self.assertEqual(mock_added_lines.return_value, actual)
        self.assertEqual({'added_files': 11, 'added_lines': 13}, stats)

    def test_compute_pfile_stats__excluded(self):
        stats = {'blacklisted_files': 10}
        config = Config(None)
        config.exclude_paths = ['tmp/*']
        farcy = self._farcy_instance(config=config)
        self.assertEqual(None, farcy._compute_pfile_stats(
            mockpfile(filename='tmp/foo'), stats))
        self.assertEqual({'blacklisted_files': 11}, stats)

    @patch('farcy.added_lines')
    def test_compute_pfile_stats__modified(self, mock_added_lines):
        mock_added_lines.return_value = {1: 1, 2: 2}
        for status in ['modified', 'renamed']:
            stats = {'modified_files': 10, 'modified_lines': 10}
            actual = self._farcy_instance()._compute_pfile_stats(
                mockpfile(patch='', status=status), stats)
            self.assertTrue(mock_added_lines.called)
            mock_added_lines.reset_mock()
            self.assertEqual(mock_added_lines.return_value, actual)
            self.assertEqual({'modified_files': 11, 'modified_lines': 12},
                             stats)

    def test_compute_pfile_stats__no_change(self):
        stats = {'unchanged_files': 10}
        self.assertEqual(None, self._farcy_instance()._compute_pfile_stats(
            mockpfile(status='added'), stats))
        self.assertEqual({'unchanged_files': 11}, stats)

    def test_compute_pfile_stats__removed(self):
        stats = {'deleted_files': 10}
        config = Config(None)
        config.exclude_paths = ['tmp/*']
        farcy = self._farcy_instance(config=config)
        self.assertEqual(None, farcy._compute_pfile_stats(
            mockpfile(filename='a/tmp/b', status='removed'), stats))
        self.assertEqual({'deleted_files': 11}, stats)

    def test_compute_pfile_stats__unexpected_status(self):
        stats = {}
        with patch.object(self.logger, 'critical') as mock_critical:
            self.assertEqual(None, self._farcy_instance()._compute_pfile_stats(
                mockpfile(patch='', status='foobar'), stats))
            self.assertTrue(mock_critical.called)
        self.assertEqual({}, stats)

    def test_get_issues__simple_module(self):
        farcy = self._farcy_instance()
        pfile = mockpfile(contents=lambda: MockInfo(decoded=b'"""A."""\n'),
                          filename='a.py')
        self.assertEqual({}, farcy.get_issues(pfile))

    def test_get_issues__no_handlers(self):
        farcy = self._farcy_instance()
        self.assertEqual({}, farcy.get_issues(mockpfile(filename='')))


class FarcyHandlePrTest(FarcyBaseTest):
    DUMMY_COMMENT = Struct(body='_[farcy \n* MatchingError', path='DummyFile',
                           position=16)

    @patch('farcy.Farcy.get_issues')
    @patch('farcy.added_lines')
    def test_handle_pr__exception_from_get_issues(self, mock_added_lines,
                                                  mock_get_issues):
        def side_effect():
            raise Exception()

        mock_added_lines.return_value = {16: 16}
        mock_get_issues.side_effect = side_effect

        pr = MagicMock(number=180, state='open', user=Struct(login='Dummy'))
        pr.commits.return_value = [Struct(sha='dummy')]
        pfile = mockpfile(patch='', status='added')
        pr.files.return_value = [pfile]

        farcy = self._farcy_instance()
        with patch.object(self.logger, 'info') as mock_info:
            farcy.handle_pr(pr)
            assert_calls(mock_info, call('Handling PR#180 by Dummy'),
                         call('PR#180 STATUS: encountered an exception in '
                              'handler. Check log.'))

        mock_added_lines.assert_called_with('')
        mock_get_issues.assert_called_once_with(pfile)
        assert_calls(farcy.repo.create_status,
                     call('dummy', 'pending', context='farcy',
                          description='started investigation'),
                     call('dummy', 'error', context='farcy',
                          description=('encountered an exception in handler. '
                                       'Check log.')))

    def test_handle_pr__pr_closed(self):
        pr = MagicMock(number=180, state='closed')
        farcy = self._farcy_instance()
        with patch.object(self.logger, 'debug') as mock_debug:
            farcy.handle_pr(pr)
            mock_debug.assert_called_with(
                'Skipping PR#180: invalid state (closed)')
        pr.refresh.assert_called_with()

    @patch('farcy.Farcy.get_issues')
    @patch('farcy.added_lines')
    def test_handle_pr__single_failure(self, mock_added_lines,
                                       mock_get_issues):
        mock_added_lines.return_value = {16: 16}
        mock_get_issues.return_value = {16: ['Dummy Failure']}

        pr = MagicMock(number=180, state='open', user=Struct(login='Dummy'))
        pr.commits.return_value = [Struct(sha='dummy')]
        pfile = mockpfile(filename='DummyFile', patch='', status='added')
        pr.files.return_value = [pfile]

        farcy = self._farcy_instance()
        with patch.object(self.logger, 'info') as mock_info:
            farcy.handle_pr(pr)
            assert_calls(mock_info,
                         call('Handling PR#180 by Dummy'),
                         call('PR#180 STATUS: found 1 issue'))

        mock_added_lines.assert_called_with('')
        mock_get_issues.assert_called_once_with(pfile)
        assert_calls(pr.create_review_comment, call(
            '{0}\n* Dummy Failure'.format(FARCY_COMMENT_START),
            'dummy', 'DummyFile', 16))
        assert_status(farcy, failures=1)

    @patch('farcy.Farcy.get_issues')
    @patch('farcy.added_lines')
    def test_handle_pr__single_failure__limit_exceeded(self, mock_added_lines,
                                                       mock_get_issues):
        mock_added_lines.return_value = {16: 16}
        mock_get_issues.return_value = {16: ['Dummy Failure']}

        pr = MagicMock(number=180, state='open', user=Struct(login='Dummy'))
        pr.commits.return_value = [Struct(sha='dummy')]
        pr.review_comments.return_value = [self.DUMMY_COMMENT] * 128

        pfile = mockpfile(filename='DummyFile', patch='', status='added')
        pr.files.return_value = [pfile]

        farcy = self._farcy_instance()
        with patch.object(self.logger, 'debug') as mock_debug:
            with patch.object(self.logger, 'info') as mock_info:
                farcy.handle_pr(pr)
                assert_calls(mock_info,
                             call('Handling PR#180 by Dummy'),
                             call('PR#180 STATUS: found 1 issue'))
            assert_calls(mock_debug,
                         call('PR#180      added_files: 1'),
                         call('PR#180      added_lines: 1'),
                         call('PR#180           issues: 1'),
                         call('PR#180   skipped_issues: 1'))

        mock_added_lines.assert_called_with('')
        mock_get_issues.assert_called_once_with(pfile)
        assert_calls(pr.create_review_comment)
        assert_status(farcy, failures=1)

    @patch('farcy.Farcy.get_issues')
    @patch('farcy.added_lines')
    def test_handle_pr__success(self, mock_added_lines, mock_get_issues):
        mock_added_lines.return_value = {16: 16}
        mock_get_issues.return_value = {3: ['Failure on non-modified line.']}

        pr = MagicMock(number=180, state='open', user=Struct(login='Dummy'))
        pr.commits.return_value = [Struct(sha='dummy')]
        pfile = mockpfile(patch='', status='added')
        pr.files.return_value = [pfile]

        farcy = self._farcy_instance()
        with patch.object(self.logger, 'info') as mock_info:
            farcy.handle_pr(pr)
            assert_calls(mock_info,
                         call('Handling PR#180 by Dummy'),
                         call('PR#180 STATUS: approves! Dummy Approval!'))

        mock_added_lines.assert_called_with('')
        mock_get_issues.assert_called_once_with(pfile)
        assert_calls(pr.create_review_comment)
        assert_status(farcy)

    def test_handle_pr__success_without_any_changed_files(self):
        pr = MagicMock(number=180, state='open', user=Struct(login='Dummy'))
        pr.commits.return_value = [Struct(sha='dummy')]
        pr.files.return_value = [mockpfile()]
        farcy = self._farcy_instance()
        with patch.object(self.logger, 'info') as mock_info:
            farcy.handle_pr(pr)
            assert_calls(mock_info,
                         call('Handling PR#180 by Dummy'),
                         call('PR#180 STATUS: approves! Dummy Approval!'))
        assert_status(farcy)

    def test_handle_pr__success_without_files(self):
        pr = MagicMock(number=180, state='open', user=Struct(login='Dummy'))
        pr.commits.return_value = [Struct(sha='dummy')]
        farcy = self._farcy_instance()
        with patch.object(self.logger, 'info') as mock_info:
            farcy.handle_pr(pr)
            assert_calls(mock_info,
                         call('Handling PR#180 by Dummy'),
                         call('PR#180 STATUS: approves! Dummy Approval!'))
        assert_status(farcy)

    def test_handle_pr__user_blacklisted(self):
        pr = Struct(number=180, user=Struct(login='Dummy'))
        farcy = self._farcy_instance()
        farcy.config.exclude_users = ['dummy']
        with patch.object(self.logger, 'debug') as mock_debug:
            farcy.handle_pr(pr)
            mock_debug.assert_called_with(
                'Skipping PR#180: Dummy is not allowed')

    def test_handle_pr__user_not_whitelisted(self):
        pr = Struct(number=180, user=Struct(login='Dummy'))
        farcy = self._farcy_instance()
        farcy.config.limit_users = ['bboe']
        with patch.object(self.logger, 'debug') as mock_debug:
            farcy.handle_pr(pr)
            mock_debug.assert_called_with(
                'Skipping PR#180: Dummy is not allowed')


class FarcyEventCallbackTest(FarcyBaseTest):
    @patch('farcy.Farcy.handle_pr')
    def test_PullRequestEvent__closed_existing(self, mock_handle_pr):
        instance = self._farcy_instance()
        instance.open_prs = {'DUMMY_BRANCH': None}

        pull_request = Struct(head=Struct(ref='DUMMY_BRANCH'), number=1337)
        event = Struct(payload={'action': 'closed',
                                'pull_request': pull_request})

        instance.PullRequestEvent(event)
        self.assertEqual({}, instance.open_prs)
        self.assertFalse(mock_handle_pr.called)

    @patch('farcy.Farcy.handle_pr')
    def test_PullRequestEvent__closed_non_existing(self, mock_handle_pr):
        instance = self._farcy_instance()
        instance.log = MagicMock()
        self.assertEqual({}, instance.open_prs)

        pull_request = Struct(head=Struct(ref='DUMMY_BRANCH'), number=1337)
        event = Struct(payload={'action': 'closed',
                                'pull_request': pull_request})

        instance.PullRequestEvent(event)
        self.assertEqual({}, instance.open_prs)
        self.assertFalse(mock_handle_pr.called)
        self.assertTrue(instance.log.warning.called)

    @patch('farcy.Farcy.handle_pr')
    def test_PullRequestEvent__opened(self, mock_handle_pr):
        instance = self._farcy_instance()
        self.assertEqual({}, instance.open_prs)

        pull_request = Struct(head=Struct(ref='DUMMY_BRANCH'), number=1337)
        event = Struct(payload={'action': 'opened',
                                'pull_request': pull_request})

        instance.PullRequestEvent(event)
        self.assertEqual({'DUMMY_BRANCH': pull_request}, instance.open_prs)
        self.assertTrue(mock_handle_pr.called)

    @patch('farcy.Farcy.handle_pr')
    def test_PullRequestEvent__reopened(self, mock_handle_pr):
        instance = self._farcy_instance()
        self.assertEqual({}, instance.open_prs)

        pull_request = Struct(head=Struct(ref='DUMMY_BRANCH'), number=1337)
        event = Struct(payload={'action': 'reopened',
                                'pull_request': pull_request})

        instance.PullRequestEvent(event)
        self.assertEqual({'DUMMY_BRANCH': pull_request}, instance.open_prs)
        self.assertFalse(mock_handle_pr.called)

    @patch('farcy.Farcy.handle_pr')
    def test_PushEvent__pr_does_not_exist(self, mock_handle_pr):
        event = Struct(payload={'ref': 'refs/heads/DUMMY_BRANCH'})
        self._farcy_instance().PushEvent(event)
        self.assertFalse(mock_handle_pr.called)

    @patch('farcy.Farcy.handle_pr')
    def test_PushEvent__pr_exists(self, mock_handle_pr):
        instance = self._farcy_instance()
        instance.open_prs['DUMMY_BRANCH'] = 0xDEADBEEF
        instance.PushEvent(Struct(payload={'ref': 'refs/heads/DUMMY_BRANCH'}))
        mock_handle_pr.assert_called_with(0xDEADBEEF)


class FarcyEventTest(FarcyBaseTest):
    def test_event_loop__ignore_events_before_start(self):
        event = Struct(actor=Struct(login=None), type='PushEvent',
                       created_at=datetime.now(UTC()), id=1)
        farcy = self._farcy_instance()  # Must have its time set second.

        events = []
        newest_id = farcy._event_loop([event], events)
        self.assertEqual(None, newest_id)
        self.assertEqual([], events)

    def test_event_loop__ignore_old_events(self):
        farcy = self._farcy_instance()  # Must have its time set first.
        farcy.last_event_id = 1

        event = Struct(actor=Struct(login=None), type='PushEvent',
                       created_at=datetime.now(UTC()), id=1)

        events = []
        newest_id = farcy._event_loop([event], events)
        self.assertEqual(None, newest_id)
        self.assertEqual([], events)

    def test_event_loop__multiple_events(self):
        farcy = self._farcy_instance()  # Must have its time set first.

        event_1 = Struct(actor=Struct(login=None), type='PushEvent',
                         created_at=datetime.now(UTC()), id=1)
        event_2 = Struct(actor=Struct(login=None), type='ForkEvent',
                         created_at=datetime.now(UTC()), id=2)
        event_3 = Struct(actor=Struct(login=None), type='PullRequestEvent',
                         created_at=datetime.now(UTC()), id=3)
        event_4 = Struct(actor=Struct(login=None), type='MemberEvent',
                         created_at=datetime.now(UTC()), id=4)

        events = []
        newest_id = farcy._event_loop([event_4, event_3, event_2, event_1],
                                      events)
        self.assertEqual(4, newest_id)
        self.assertEqual([event_1, event_3], events)

    def test_event_loop__no_events(self):
        events = []
        newest_id = self._farcy_instance()._event_loop([], events)
        self.assertEqual(None, newest_id)
        self.assertEqual([], events)

    def test_event_loop__single_event(self):
        farcy = self._farcy_instance()  # Must have its time set first.
        event = Struct(actor=Struct(login=None), type='PushEvent',
                       created_at=datetime.now(UTC()), id=0xDEADBEEF)

        events = []
        newest_id = farcy._event_loop([event], events)
        self.assertEqual(0xDEADBEEF, newest_id)
        self.assertEqual([event], events)

    def test_events__end_loop(self):
        farcy = self._farcy_instance()
        self.assertEqual(None, farcy.last_event_id)

        event = Struct(actor=Struct(login=None), type='PushEvent',
                       created_at=datetime.now(UTC()), id=0xDEADBEEF)
        farcy.repo.events.return_value = Struct(
            [event], etag='DUMMY_ETAG',
            last_response=Struct(headers={'X-Poll-Interval': 100}))

        event_itr = farcy.events()
        self.assertEqual(event, next(event_itr))
        farcy.running = False
        self.assertRaises(StopIteration, next, event_itr)

    @patch('farcy.Farcy._event_loop')
    @patch('time.sleep')
    def test_events__network_exception(self, mock_sleep, mock_event_loop):
        farcy = self._farcy_instance()
        event = Struct(actor=Struct(login=None), type='PushEvent',
                       created_at=datetime.now(UTC()), id=0xDEADBEEF)

        call_count = [0]

        def side_effect(_, events):
            call_count[0] += 1
            if call_count[0] == 1:
                raise ConnectionError('Foo')
            else:
                events.append(event)
                return event.id

        mock_event_loop.side_effect = side_effect
        self.assertEqual(event, next(farcy.events()))
        self.assertEqual(0xDEADBEEF, farcy.last_event_id)
        self.assertTrue(mock_sleep.called_with(1))

    def test_events__prevent_duplicate_calls(self):
        farcy = self._farcy_instance()
        self.assertEqual(None, farcy.last_event_id)

        event = Struct(actor=Struct(login=None), type='PushEvent',
                       created_at=datetime.now(UTC()), id=0xDEADBEEF)
        farcy.repo.events.return_value = Struct([event], etag='DUMMY_ETAG')

        self.assertEqual(event, next(farcy.events()))
        self.assertEqual(0xDEADBEEF, farcy.last_event_id)

        self.assertRaises(FarcyException, next, farcy.events())

    @patch('farcy.Farcy.events')
    @patch('farcy.Farcy.PushEvent')
    def test_run(self, mock_callback, mock_events):
        event1 = Struct(type='PushEvent', uniq=1)
        event2 = Struct(type='PushEvent', uniq=2)
        self.assertEqual(event1, event1)
        self.assertNotEqual(event1, event2)

        mock_events.return_value = [event1, event2]

        self._farcy_instance().run()
        assert_calls(mock_callback, call(event1), call(event2))
        mock_callback.assert_called_with(event2)

    @patch('farcy.Farcy.handle_pr')
    def test_run__single_pull_request(self, mock_handle_pr):
        farcy = self._farcy_instance()
        farcy.repo.pull_request.side_effect = lambda x: x
        farcy.config.pull_requests = '418'
        farcy.run()
        assert_calls(mock_handle_pr, call(418, force=True))

    @patch('farcy.Farcy.handle_pr')
    def test_run__multiple_pull_requests(self, mock_handle_pr):
        farcy = self._farcy_instance()
        farcy.repo.pull_request.side_effect = lambda x: x
        farcy.config.pull_requests = '360,180,720'
        farcy.run()
        assert_calls(mock_handle_pr, call(180, force=True),
                     call(360, force=True), call(720, force=True))


class MainTest(unittest.TestCase):
    @patch('farcy.Farcy')
    @patch('farcy.Config')
    def test_main__farcy_exception_in_run(self, mock_config, mock_farcy):
        def side_effect():
            raise FarcyException
        mock_farcy.return_value.run.side_effect = side_effect
        self.assertEqual(1, main())

    @patch('farcy.Farcy')
    @patch('farcy.Config')
    def test_main__keyboard_interrupt_in_farcy(self, mock_config, mock_farcy):
        def side_effect(_):
            raise KeyboardInterrupt
        mock_farcy.side_effect = side_effect
        self.assertEqual(0, main())

    @patch('farcy.Farcy')
    @patch('farcy.Config')
    def test_main__keyboard_interrupt_in_run(self, mock_config, mock_farcy):
        def side_effect():
            raise KeyboardInterrupt
        mock_farcy.return_value.run.side_effect = side_effect
        self.assertEqual(0, main())

    @patch('farcy.Config')
    def test_main__no_repo_specified(self, mock_config):
        mock_config.return_value.repository = None
        self.assertEqual(2, main())

    @patch('farcy.Farcy')
    @patch('farcy.Config')
    def test_main__no_exception(self, mock_config, mock_farcy):
        self.assertEqual(None, main())
        self.assertTrue(mock_config.called)
        self.assertTrue(mock_farcy.called)
        self.assertTrue(mock_farcy.return_value.run.called)


class NoHandlerDebugFactory(unittest.TestCase):
    def setUp(self):
        self.farcy = MagicMock()

    def test_no_handler_factory__cache_response(self):
        func = no_handler_debug_factory(1)
        func(self.farcy, '.js')
        func(self.farcy, '.js')
        self.farcy.log.debug.assert_called_once_with(
            'No handlers for extension .js')

    def test_no_handler_factory__output_when_cache_expired(self):
        func = no_handler_debug_factory(0)
        func(self.farcy, '.js')
        func(self.farcy, '.js')
        calls = [call('No handlers for extension .js')] * 2
        assert_calls(self.farcy.log.debug, *calls)

    def test_no_handler_factory__multiple_calls(self):
        func = no_handler_debug_factory(1)
        func(self.farcy, '.js')
        func(self.farcy, '.css')
        assert_calls(self.farcy.log.debug,
                     call('No handlers for extension .js'),
                     call('No handlers for extension .css'))



"""Farcy, a code review bot for github pull requests.

Usage: farcy.py [-D | --logging=LEVEL] [--comments-per-pr=LIMIT]
                [--exclude-path=PATTERN...]
                [--limit-user=USER...] [options] [REPOSITORY]

Options:

  -s ID, --start=ID  The event id to start handling events from.
  -p ID, --pr=ID     Process only the provided pull request(s).
  -D, --debug        Enable debugging mode. Enables all logging output
                     and prevents the posting of comments.
  --logging=LEVEL    Specify the log level* to output.
  -h, --help         Show this screen.
  --version          Show the program's version.
  -X PATTERN, --exclude-path=PATTERN  Exclude paths that match pattern
                                      (npm_modules/*).
  -u USER, --limit-user=USER          Limit processed pull requests to pull
                                      requests created by USER. This argument
                                      can be provided multiple times, and each
                                      USER token can contain a comma separated
                                      list of users.
  -C LIMIT, --comments-per-pr=LIMIT   Maximum number of comments added by
                                      Farcy per pull request.

* Available log levels:
    https://docs.python.org/3/library/logging.html#logging-levels

"""

from __future__ import print_function
from collections import Counter, defaultdict
from datetime import datetime
from docopt import docopt
from fnmatch import fnmatch
from github3.exceptions import ServerError, UnprocessableEntity
from random import choice
from requests import ConnectionError
from shutil import rmtree
from tempfile import mkdtemp
from timeit import default_timer
from update_checker import UpdateChecker
import logging
import os
import sys
import time
from .const import (__version__, APPROVAL_PHRASES, FARCY_COMMENT_START,
                    STATUS_CONTEXT, VERSION_STR)
from .exceptions import FarcyException, HandlerException
from .helpers import added_lines, plural
from .objects import Config, ErrorTracker, UTC


def no_handler_debug_factory(duration=3600):
    """Return a function to cache 'No handler for...' messages for an hour."""
    last_logged = {}

    def log(obj, ext):
        now = default_timer()
        if now - last_logged.get(ext, 0) > duration:
            obj.log.debug('No handlers for extension {0}'.format(ext))
        last_logged[ext] = now
    return log


class Farcy(object):
    """A bot to automate some code-review processes on GitHub pull requests."""

    EVENTS = {'PullRequestEvent', 'PushEvent'}
    _update_checked = False

    def __init__(self, config):
        """Initialize an instance of Farcy that monitors owner/repository."""
        # Configure logging
        self.config = config
        self.log = logging.getLogger(__name__)
        self.log.setLevel(config.log_level_int)

        # Prepare logging
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(
            '%(asctime)s %(levelname)8s %(message)s', '%Y/%m/%d %H:%M:%S'))
        self.log.addHandler(handler)
        self.log.info('Logging enabled at level {0}'.format(config.log_level))

        if config.start_event:
            self.start_time = None
            self.last_event_id = int(config.start_event) - 1
        else:
            self.start_time = datetime.now(UTC())
            self.last_event_id = None

        self._load_handlers()

        # Initialize the repository to monitor
        self.repo = config.session.repository(
            *self.config.repository.split('/'))
        if self.repo is None:
            raise FarcyException('Invalid owner or repository name: {0}'
                                 .format(self.config.repository))
        # Keep track of open pull requests
        self.open_prs = {}
        for pr in self.repo.pull_requests(state='open'):
            self.open_prs[pr.head.ref] = pr

        # Check for farcy package updates
        if not self._update_checked:
            result = UpdateChecker().check(__name__, __version__)
            if result:
                self.log.info(result)
            self._update_checked = True

        self.running = False

    def _compute_pfile_stats(self, pfile, stats):
        added = None
        if self.config.exclude_paths is not None and \
                any(fnmatch(pfile.filename, pattern) for pattern
                    in self.config.exclude_paths):
            stats['blacklisted_files'] += 1
        elif pfile.status == 'removed':  # Ignore deleted files
            stats['deleted_files'] += 1
        elif pfile.patch is None:  # Ignore files without changes
            stats['unchanged_files'] += 1
        elif pfile.status in ('modified', 'renamed'):
            # Only report issues on the changed lines
            added = added_lines(pfile.patch)
            stats['modified_files'] += 1
            stats['modified_lines'] += len(added)
        elif pfile.status == 'added':
            added = added_lines(pfile.patch)
            stats['added_files'] += 1
            stats['added_lines'] += len(added)
        else:
            self.log.critical('Unexpected file status {0} on {1}'
                              .format(pfile.status, pfile.filename))
        return added

    def _event_loop(self, itr, events):
        newest_id = None
        for event in itr:
            # Stop when we've already seen something
            if self.last_event_id and int(event.id) <= self.last_event_id \
               or self.start_time and event.created_at < self.start_time:
                break

            self.log.debug('EVENT {eid} {time} {etype} {user}'.format(
                eid=event.id, time=event.created_at, etype=event.type,
                user=event.actor.login))
            newest_id = newest_id or int(event.id)

            # Add relevent events in reverse order
            if event.type in self.EVENTS:
                events.insert(0, event)
        return newest_id

    def _fail_allowed(self, pr):
        if self.config.user_allowed(pr.user.login):
            return None
        return ('Skipping PR#{0}: {1} is not allowed'
                .format(pr.number, pr.user.login))

    def _fail_closed(self, pr):
        pr.refresh()
        if pr.state == 'open':
            return None
        return ('Skipping PR#{0}: invalid state ({1})'
                .format(pr.number, pr.state))

    def _get_state(self, issues, exception):
        if exception:
            return 'error', 'encountered an exception in handler. Check log.'
        if issues > 0:
            return 'failure', 'found {0}'.format(plural(issues, 'issue'))
        return 'success', 'approves! {0}!'.format(choice(APPROVAL_PHRASES))

    def _handle_pr_file(self, pfile, pr, sha, data):
        """Return whether or not an exception occured."""
        added = self._compute_pfile_stats(pfile, data['stats'])
        if added is None:
            return False

        try:
            file_issues = self.get_issues(pfile)
        except Exception:
            self.log.exception('Failure with get_issues for {0}'
                               .format(pfile.filename))
            return True

        for line, messages in file_issues.items():
            if line not in added:  # Skip unadded/unmodified lines.
                continue
            for message in messages:
                data['errors'].track(message, pfile.filename, added[line])

        exception_occurred = False
        for line, violations in data['errors'].errors(pfile.filename):
            if data['comments'] >= self.config.pr_issue_report_limit:
                data['stats']['skipped_issues'] += 1
                continue

            if self.config.debug:
                # Only log each issue if we're in debugging mode because we
                # don't want the logs in non-debugging mode to be noisy.
                self.log.info('PR#{0} ({1}:{2}): {3}"'.format(
                    pr.number, pfile.filename, line, violations))
            else:
                msg = '\n'.join(
                    [FARCY_COMMENT_START] + ['* {}'.format(violation)
                                             for violation in violations])
                try:
                    (pr.create_review_comment(msg, sha, pfile.filename, line)
                     .html_url['href'])
                except UnprocessableEntity as exc:
                    self.log.exception('Failure with create_review_comment for'
                                       ' {0} on line {1}'
                                       .format(pfile.filename, line))
                    self.log.exception(str(exc))
                    exception_occurred = True

            # `data['comments']` is misleading when in debug mode.  What
            # it really means is the number of comments that would be on
            # on the pr (existing + new) when not in debug mode.
            data['comments'] += 1
        return exception_occurred

    def _load_handlers(self):
        from . import handlers
        self._ext_to_handler = defaultdict(list)
        active = []
        for handler in (handlers.ESLint, handlers.Flake8, handlers.Pep257,
                        handlers.Rubocop, handlers.SCSSLint):
            try:
                handler_inst = handler()
            except HandlerException:
                continue
            for ext in handler.EXTENSIONS:
                self._ext_to_handler[ext].append(handler_inst)
            active.append(handler_inst.name)
        if active:
            self.log.info('Active handlers: %s', ', '.join(active))
        else:
            self.log.warning('No active handlers')

    def _set_status(self, sha, status, description):
        if not self.config.debug:
            self.repo.create_status(sha, status, context=STATUS_CONTEXT,
                                    description=description)

    def events(self):
        """Yield repository events in order."""
        if self.running:
            raise FarcyException('Can only enter `events` once.')

        etag = None
        sleep_time = None  # This value will be overwritten.
        self.running = True
        while self.running:
            if sleep_time:  # Only sleep before we're about to make requests.
                time.sleep(sleep_time)

            # Fetch events
            events = []
            itr = self.repo.events(etag=etag)
            try:
                newest_id = self._event_loop(itr, events)
            except (ConnectionError, ServerError) as exc:
                self.log.error('Error in event generation loop: {0}'
                               .format(exc))
                sleep_time = 1
                continue

            etag = itr.etag
            self.last_event_id = newest_id or self.last_event_id

            # Yield events from oldest to newest
            for event in events:
                yield event

            sleep_time = int(itr.last_response.headers.get('X-Poll-Interval',
                                                           sleep_time))

    def get_issues(self, pfile):
        """Return a dictionary of issues for the file."""
        ext = os.path.splitext(pfile.filename)[1]
        handlers = self._ext_to_handler.get(ext)
        if not handlers:  # Do nothing if there are no handlers
            self.no_handler_debug(ext)
            return {}
        retval = {}

        try:
            tmpdir = mkdtemp()
            filepath = os.path.join(tmpdir, os.path.basename(pfile.filename))
            with open(filepath, 'wb') as fp:
                fp.write(pfile.contents().decoded)
            for handler in handlers:
                retval.update(handler.process(fp.name))
        finally:
            rmtree(tmpdir)

        return retval

    def handle_pr(self, pr, force=False):
        """Provide code review on pull request."""
        failure = not force and (self._fail_allowed(pr) or
                                 self._fail_closed(pr))
        if failure:
            self.log.debug(failure)
            return

        sha = list(pr.commits())[-1].sha
        self._set_status(sha, 'pending', 'started investigation')
        self.log.info('Handling PR#{0} by {1}'
                      .format(pr.number, pr.user.login))

        exception = False
        error_tracker = ErrorTracker(pr.review_comments(),
                                     self.config.comment_group_threshold)
        handle_data = {'comments': error_tracker.github_message_count,
                       'errors': error_tracker,
                       'stats': Counter()}
        for pfile in pr.files():
            exception = self._handle_pr_file(
                pfile, pr, sha, handle_data) or exception

        handle_data['stats']['issues'] += error_tracker.new_issue_count
        handle_data['stats']['hidden'] += error_tracker.hidden_issue_count

        # Log the statistics for the PR
        for key, count in sorted(handle_data['stats'].items()):
            if count > 0:
                self.log.debug('PR#{0} {1:>16}: {2}'
                               .format(pr.number, key, count))

        state, message = self._get_state(handle_data['stats']['issues'],
                                         exception)
        self._set_status(sha, state, message)
        self.log.info('PR#{0} STATUS: {1}'.format(pr.number, message))

    no_handler_debug = no_handler_debug_factory()

    def PullRequestEvent(self, event):
        """Check commits on new pull requests."""
        pr = event.payload['pull_request']
        action = event.payload['action']
        self.log.debug('PullRequest #{num} {action} on branch {branch}'
                       .format(action=action, branch=pr.head.ref,
                               num=pr.number))
        if action == 'closed':
            if pr.head.ref in self.open_prs:
                del self.open_prs[pr.head.ref]
            else:
                self.log.warning('open_prs did not contain {0}'
                                 .format(pr.head.ref))
        elif action == 'opened':
            self.open_prs[pr.head.ref] = pr
            self.handle_pr(pr)
        elif action == 'reopened':
            self.open_prs[pr.head.ref] = pr

    def PushEvent(self, event):
        """Check push commits only to open pull requests."""
        ref = event.payload['ref']
        assert ref.startswith('refs/heads/')
        pull_request = self.open_prs.get(ref.rsplit('/', 1)[1])
        if pull_request:
            self.handle_pr(pull_request)

    def run(self):
        """Run the bot until ctrl+c is received."""
        if self.config.pull_requests is not None:
            for number in sorted(int(x) for x in self.config.pull_requests):
                self.handle_pr(self.repo.pull_request(number), force=True)
            return

        self.log.info('Monitoring {0}'.format(self.repo.html_url))
        for event in self.events():
            attempts = 3
            while attempts > 0:
                if attempts < 3:  # Sleep only on subsequent attempts.
                    time.sleep(4 ** (3 - attempts))
                try:
                    getattr(self, event.type)(event)
                    attempts = 0
                except Exception as exc:
                    attempts -= 1
                    self.log.error('Error with event ({0}): {1}'
                                   .format(event, exc))
                    self.log.info('Retrying {0} more time(s).'
                                  .format(attempts))


def main():
    """Provide an entry point into Farcy."""
    args = docopt(__doc__, version=VERSION_STR)
    config = Config(args['REPOSITORY'], debug=args['--debug'],
                    exclude_paths=args['--exclude-path'],
                    limit_users=args['--limit-user'],
                    log_level=args['--logging'],
                    pr_issue_report_limit=args['--comments-per-pr'],
                    pull_requests=args['--pr'],
                    start_event=args['--start'])
    if config.repository is None:
        sys.stderr.write('No repository specified\n')
        return 2

    try:
        Farcy(config).run()
    except KeyboardInterrupt:
        sys.stderr.write('Farcy shutting down. Goodbye!\n')
        return 0
    except FarcyException as exc:
        sys.stderr.write('{0}\n'.format(exc))
        return 1



"""Farcy setup.py."""

import os
import re
from setuptools import setup

PACKAGE_NAME = 'farcy'

HERE = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(HERE, 'README.rst')) as fp:
    README = fp.read()
with open(os.path.join(HERE, PACKAGE_NAME, 'const.py')) as fp:
    VERSION = re.search("__version__ = '([^']+)'", fp.read()).group(1)


setup(name=PACKAGE_NAME,
      author='Bryce Boe',
      author_email='bbzbryce@gmail.com',
      classifiers=['Intended Audience :: Developers',
                   'License :: OSI Approved :: BSD License',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3.3',
                   'Programming Language :: Python :: 3.4',
                   'Programming Language :: Python :: 3.5'],
      description='A code review bot for github pull requests',
      entry_points={'console_scripts':
                    ['{0} = {0}:main'.format(PACKAGE_NAME)]},
      extras_require={'python': ['flake8 >= 2.2.5', 'pep257 >= 0.3.2']},
      install_requires=['botocore >= 0.74.0',
                        'docopt >= 0.6.2',
                        'github3.py >= 1.0.0a2',
                        'update_checker >= 0.11'],
      keywords=['code review', 'pull request'],
      license='Simplified BSD License',
      long_description=README,
      packages=[PACKAGE_NAME],
      tests_require=['mock >= 1.0.1',
                     # https://bugs.launchpad.net/pbr/+bug/1493735
                     'pbr < 1.7.0'],
      test_suite='test',
      url='https://github.com/appfolio/farcy',
      version=VERSION)



"""Defines the standard objects used by Farcy."""

try:
    from configparser import ConfigParser  # PY3
except ImportError:
    from ConfigParser import SafeConfigParser as ConfigParser  # PY2

from datetime import timedelta, tzinfo
import logging
import os
import re
from .const import CONFIG_DIR, FARCY_COMMENT_START
from .exceptions import FarcyException
from .helpers import get_session, parse_bool, parse_set


class Config(object):
    """Holds configuration for Farcy."""

    ATTRIBUTES = {'comment_group_threshold', 'debug', 'exclude_paths',
                  'exclude_users', 'limit_users', 'log_level',
                  'pr_issue_report_limit', 'pull_requests', 'start_event'}
    INT_ATTRS = {'comment_group_threshold', 'pr_issue_report_limit',
                 'start_event'}
    LOG_LEVELS = {'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'}
    PATH = os.path.join(CONFIG_DIR, 'farcy.conf')

    @property
    def log_level_int(self):
        """Int value of the log level."""
        return getattr(logging, self.log_level)

    @property
    def session(self):
        """Return GitHub session. Create if necessary."""
        if self._session is None:
            self._session = get_session()
        return self._session

    def __init__(self, repository, **overrides):
        """Initialize a config with default values."""
        self._session = None
        self.repository = repository
        self.set_defaults()
        self.load_config_file()
        self.override(**overrides)

    def __repr__(self):
        """String representation of the config."""
        keys = sorted(x for x in self.__dict__ if not x.startswith('_') and
                      x != 'repository')
        arg_fmt = ', '.join(['{0}={1!r}'.format(key, getattr(self, key))
                             for key in keys])
        return 'Config({0!r}, {1})'.format(self.repository, arg_fmt)

    def __setattr__(self, attr, value):
        """
        Set new config attribute.

        Validates new attribute values and tracks if changed from default.

        """
        if attr == 'debug' and parse_bool(value):
            # Force log level when in debug mode
            setattr(self, 'log_level', 'DEBUG')
        elif attr in ('exclude_paths', 'pull_requests'):
            if value is not None:
                value = parse_set(value)
        elif attr in ('exclude_users', 'limit_users'):
            if value:
                value = parse_set(value, normalize=True)
        elif attr == 'log_level' and self.debug:
            return  # Don't change level in debug mode
        elif attr == 'log_level' and value is not None:
            value = value.upper()
            if value not in self.LOG_LEVELS:
                raise FarcyException('Invalid log level: {0}'.format(value))
        elif attr == 'repository' and value is not None:
            repo_parts = value.split('/')
            if len(repo_parts) != 2:
                raise FarcyException('Invalid repository: {0}'.format(value))
        elif attr in self.INT_ATTRS:
            if value is not None:
                value = int(value)
        super(Config, self).__setattr__(attr, value)
        if getattr(self, 'exclude_users', None) and \
           getattr(self, 'limit_users', None):
            raise FarcyException('Either exclude_users or limit_users '
                                 'can be provided, but not both.')

    def load_config_file(self):
        """Load value overrides from configuration file."""
        if not os.path.isfile(self.PATH):
            return

        config_file = ConfigParser()
        config_file.read(self.PATH)

        if not self.repository and \
                config_file.has_option('DEFAULT', 'repository'):
            self.repository = config_file.get('DEFAULT', 'repository')

        self.override(**dict(config_file.items(
            self.repository if config_file.has_section(self.repository)
            else 'DEFAULT')))

    def override(self, **overrides):
        """Override the config values passed as keyword arguments."""
        for attr, value in overrides.items():
            if attr in self.ATTRIBUTES and value:
                setattr(self, attr, value)

    def set_defaults(self):
        """Set the default config values."""
        self.comment_group_threshold = 3
        self.debug = False
        self.exclude_paths = None
        self.exclude_users = None
        self.limit_users = None
        self.log_level = 'ERROR'
        self.pr_issue_report_limit = 128
        self.pull_requests = None
        self.start_event = None

    def user_allowed(self, user):
        """Return if user is allowed."""
        if self.exclude_users:
            return user.lower() not in self.exclude_users
        elif self.limit_users:
            return user.lower() in self.limit_users
        return True


class ErrorMessage(object):
    """An error message keeps track the lines a single error appears on."""

    def __init__(self, message, group_threshold):
        """Initialize an ErrorMessage object.

        :param message: The error message.
        :param group_threshold: The number of lines across which to group
            this ErrorMessage. If the threshold is 2, and this error occurs on
            lines 10, 12, 14, 17, then the first three errors will be grouped
            together as there are at most 2 lines between any two instances
            within the group.

        """
        self.group_threshold = group_threshold
        self.groups = set()
        self.lines = {}  # Value is true when it's on github
        self.message = message

    def messages(self):
        """Yield a tuple containing (line, message).

        Messages near each other will be grouped and the message will indicate
        how many lines are covered by the message.

        """
        def output(start, count, span):
            if count > 1:
                return (start, '{0} <sub>{1}x spanning {2} lines</sub>'
                        .format(self.message, count, span + 1))
            return (start, self.message)

        start = last = None
        count = 0
        for line, skip in sorted(self.lines.items()):
            if skip:
                continue
            if start is None:
                start = last = line
            if line - last > self.group_threshold:
                if (start, count) not in self.groups:
                    yield output(start, count, last - start)
                count = 0
                start = line
            count += 1
            last = line
        if start and (start, count) not in self.groups:
            yield output(start, count, last - start)

    def track(self, line, on_github=False):
        """Track the line and return self."""
        self.lines[line] = self.lines.get(line, False) or on_github
        return self

    def track_group(self, line, count):
        """Record a grouping for this message that is on github."""
        self.groups.add((line, count))
        return self


class ErrorTracker(object):
    """Track ErrorMessages across multiple files."""

    FARCY_PREFIX = FARCY_COMMENT_START.split('v')[0]
    GROUP_MATCH = re.compile('(.+) <sub>(\d+)x spanning \d+ lines</sub>')

    @classmethod
    def _parse_group_message(cls, message):
        match = cls.GROUP_MATCH.match(message)
        return match.groups() if match else None

    def __init__(self, github_comments, group_threshold):
        """Initialize an ErrorTracker object.

        :param github_comments: The comments that already exist on github.
        :param group_threshold: The number of lines across which to group
            this ErrorMessage. If the threshold is 2, and this error occurs on
            lines 10, 12, 14, 17, then the first three errors will be grouped
            together as there are at most 2 lines between any two instances
            within the group.

        """
        self.by_file = {}
        self.github_message_count = 0
        self.group_threshold = group_threshold
        self.hidden_issue_count = 0
        self.new_issue_count = 0
        self.from_github_comments(github_comments)

    def errors(self, filename):
        """Generate tuples containing (line, [errors...])."""
        by_line = {}
        for error in self.by_file.get(filename, {}).values():
            for line, message in error.messages():
                by_line.setdefault(line, []).append(message)
        for line in sorted(by_line):
            yield (line, sorted(by_line[line]))

    def from_github_comments(self, comments):
        """Populate the error tracker with Farcy comments from github."""
        for comment in comments:
            if not comment.body.startswith(self.FARCY_PREFIX):
                continue
            if comment.position == 0:
                self.hidden_issue_count += 1
                continue
            self.github_message_count += 1
            for issue in comment.body.split('\n')[1:]:
                self.track(issue[2:], comment.path, comment.position, True)

    def track(self, message, filename, line, is_github=False):
        """Track message in filename on line."""
        parts = self._parse_group_message(message)
        if parts:
            message = parts[0]

        error_message = self.by_file.setdefault(filename, {}).setdefault(
            message, ErrorMessage(message, self.group_threshold))

        if parts:
            error_message.track_group(line, int(parts[1]))
        else:
            if not is_github:
                self.new_issue_count += 1
            error_message.track(line, is_github)


class UTC(tzinfo):
    """Provides a simple UTC timezone class.

    Source: http://docs.python.org/release/2.4.2/lib/datetime-tzinfo.html

    """

    def dst(self, dt):
        """Return the daylight saving time adjustment."""
        return timedelta(0)

    def tzname(self, dt):
        """Name of the timezone."""
        return 'UTC'

    def utcoffset(self, dt):
        """Offset from UTC time."""
        return timedelta(0)



"""Helper methods and classes."""

from github3 import GitHub
from github3.exceptions import GitHubError
import os
import sys
from .const import NUMBER_RE, CONFIG_DIR
from .exceptions import FarcyException

if sys.version_info >= (3, 0):
    basestring = str


def added_lines(patch):
    """Return a mapping of added line numbers to the patch line numbers."""
    added = {}
    lineno = None
    position = 0
    for line in patch.split('\n'):
        if line.startswith('@@'):
            lineno = int(NUMBER_RE.match(line.split('+')[1]).group(1))
        elif line.startswith(' '):
            lineno += 1
        elif line.startswith('+'):
            added[lineno] = position
            lineno += 1
        elif line == "\ No newline at end of file":
            continue
        else:
            assert line.startswith('-')
        position += 1
    return added


def ensure_config_dir():
    """Ensure Farcy config dir exists."""
    if not os.path.isdir(CONFIG_DIR):
        os.makedirs(CONFIG_DIR, mode=0o700)


def get_session():
    """Fetch and/or load API authorization token for GITHUB."""
    ensure_config_dir()
    credential_file = os.path.join(CONFIG_DIR, 'github_auth')
    if os.path.isfile(credential_file):
        with open(credential_file) as fd:
            token = fd.readline().strip()
        gh = GitHub(token=token)
        try:  # Test connection before starting
            gh.is_starred('github', 'gitignore')
            return gh
        except GitHubError as exc:
            raise_unexpected(exc.code)
            sys.stderr.write('Invalid saved credential file.\n')

    from getpass import getpass
    from github3 import authorize

    user = prompt('GITHUB Username')
    try:
        auth = authorize(
            user, getpass('Password for {0}: '.format(user)), 'repo',
            'Farcy Code Reviewer',
            two_factor_callback=lambda: prompt('Two factor token'))
    except GitHubError as exc:
        raise_unexpected(exc.code)
        raise FarcyException(exc.message)

    with open(credential_file, 'w') as fd:
        fd.write('{0}\n{1}\n'.format(auth.token, auth.id))
    return GitHub(token=auth.token)


def parse_bool(value):
    """Return whether or not value represents a True or False value."""
    if isinstance(value, basestring):
        return value.lower() in ['1', 'on', 't', 'true', 'y', 'yes']
    return bool(value)


def parse_set(item_or_items, normalize=False):
    """Return a set of unique tokens in item_or_items.

    :param item_or_items: Can either be a string, or an iterable of strings.
      Each string can contain one or more items separated by commas, these
      items will be expanded, and empty tokens will be removed.
    :param normalize: When true, lowercase all tokens.

    """
    if isinstance(item_or_items, basestring):
        item_or_items = [item_or_items]

    items = set()
    for item in item_or_items:
        for token in (x.strip() for x in item.split(',') if x.strip()):
            items.add(token.lower() if normalize else token)
    return items if items else None


def plural(items, word):
    """Return number of items followed by the right form  of ``word``.

    ``items`` can either be an int or an object whose cardinality can be
    discovered via `len(items)`.

    The plural of ``word`` is assumed to be made by adding an ``s``.

    """
    item_count = items if isinstance(items, int) else len(items)
    word = word if item_count == 1 else word + 's'
    return '{0} {1}'.format(item_count, word)


def prompt(msg):
    """Output message and return striped input."""
    sys.stdout.write('{0}: '.format(msg))
    sys.stdout.flush()
    return sys.stdin.readline().strip()


def raise_unexpected(code):
    """Called from with in an except block.

    Re-raises the exception if we don't know how to handle it.

    """
    if code != 401:
        raise



"""Farcy helpers test file."""

from __future__ import print_function
from collections import namedtuple
from farcy import helpers
from github3 import GitHub, GitHubError
from io import IOBase
from mock import MagicMock, patch
import unittest
import farcy.exceptions as exceptions


MockResponse = namedtuple('MockResponse', ['content', 'status_code'])


class MockComment(object):
    """Imitates a review comment."""

    def __init__(self, body=None, path=None, position=None, issues=None):
        if issues:
            body = '\n'.join([helpers.FARCY_COMMENT_START] +
                             ['* ' + issue for issue in issues])
        self.body = body
        self.path = path
        self.position = position


class EnsureConfigDirTest(unittest.TestCase):
    @patch('os.makedirs')
    @patch('os.path.isdir')
    def test_ensure_config_dir__create(self, mock_isdir, mock_makedirs):
        mock_isdir.return_value = False
        helpers.ensure_config_dir()
        self.assertTrue(mock_isdir.called)
        mock_makedirs.assert_called_with(helpers.CONFIG_DIR, mode=0o700)

    @patch('os.makedirs')
    @patch('os.path.isdir')
    def test_ensure_config_dir__no_create(self, mock_isdir, mock_makedirs):
        mock_isdir.return_value = True
        helpers.ensure_config_dir()
        self.assertTrue(mock_isdir.called)
        self.assertFalse(mock_makedirs.called)


class GetSessionTest(unittest.TestCase):
    @patch('farcy.helpers.open', create=True)
    @patch('github3.authorize')
    @patch('getpass.getpass')
    @patch('farcy.helpers.prompt')
    @patch('farcy.helpers.os.path')
    def test_get_session__authenticate(self, mock_path, mock_prompt,
                                       mock_getpass, mock_authorize,
                                       mock_open):
        mock_path.isfile.return_value = False
        self.assertTrue(isinstance(helpers.get_session(), GitHub))
        self.assertTrue(mock_prompt.called)
        self.assertTrue(mock_getpass.called)
        self.assertTrue(mock_open.called)

    @patch('github3.authorize')
    @patch('getpass.getpass')
    @patch('farcy.helpers.prompt')
    @patch('farcy.helpers.os.path')
    def test_get_session__authenticate_with_exceptions(
            self, mock_path, mock_prompt, mock_getpass, mock_authorize):
        mock_path.isfile.return_value = False

        mock_response = MockResponse(content='', status_code=401)
        mock_authorize.side_effect = GitHubError(mock_response)
        self.assertRaises(exceptions.FarcyException, helpers.get_session)

        self.assertTrue(mock_prompt.called)
        self.assertTrue(mock_getpass.called)

        mock_response = MockResponse(content='', status_code=101)
        mock_authorize.side_effect = GitHubError(mock_response)
        self.assertRaises(GitHubError, helpers.get_session)

        mock_authorize.side_effect = TypeError
        self.assertRaises(TypeError, helpers.get_session)

    @patch.object(GitHub, 'is_starred')
    @patch('farcy.helpers.open', create=True)
    @patch('farcy.helpers.os.path')
    def test_get_session__from_credentials_file(self, mock_path, mock_open,
                                                mock_is_starred):
        mock_path.expanduser.return_value = 'mock_path'
        mock_path.isfile.return_value = True
        mock_open.return_value = MagicMock(spec=IOBase)
        self.assertTrue(helpers.get_session())
        self.assertTrue(mock_is_starred.called)

    @patch('github3.authorize')
    @patch('getpass.getpass')
    @patch('farcy.helpers.prompt')
    @patch('farcy.helpers.sys.stderr')
    @patch.object(GitHub, 'is_starred')
    @patch('farcy.helpers.open', create=True)
    @patch('farcy.helpers.os.path')
    def test_get_session__from_credentials_file__handled_exception(
            self, mock_path, mock_open, mock_is_starred, mock_stderr,
            mock_prompt, mock_getpass, mock_authorize):
        mock_path.expanduser.return_value = 'mock_path'
        mock_path.isfile.return_value = True
        mock_open.return_value = MagicMock(spec=IOBase)

        mock_response = MockResponse(content='', status_code=401)
        mock_is_starred.side_effect = GitHubError(mock_response)
        self.assertTrue(isinstance(helpers.get_session(), GitHub))
        self.assertTrue(mock_stderr.write.called)
        self.assertTrue(mock_prompt.called)
        self.assertTrue(mock_getpass.called)
        self.assertTrue(mock_open.called)

    @patch.object(GitHub, 'is_starred')
    @patch('farcy.helpers.open', create=True)
    @patch('farcy.helpers.os.path')
    def test_get_session__from_credentials_file__unhandled_exception(
            self, mock_path, mock_open, mock_is_starred):
        mock_path.expanduser.return_value = 'mock_path'
        mock_path.isfile.return_value = True
        mock_open.return_value = MagicMock(spec=IOBase)

        mock_is_starred.side_effect = TypeError
        self.assertRaises(TypeError, helpers.get_session)


class PatchFunctionTest(unittest.TestCase):
    def test_added_lines(self):
        self.assertEqual({}, helpers.added_lines('@@+15'))
        self.assertEqual({1: 1}, helpers.added_lines('@@+1\n+wah'))
        self.assertEqual({15: 1}, helpers.added_lines('@@+15\n+wah'))
        self.assertEqual({16: 2}, helpers.added_lines('@@+15\n \n+wah'))
        self.assertEqual({1: 2}, helpers.added_lines('@@+1\n-\n+wah'))
        self.assertEqual({15: 2}, helpers.added_lines('@@+15\n-\n+wah'))
        self.assertEqual({16: 3}, helpers.added_lines('@@+15\n-\n \n+wah'))
        self.assertEqual({1: 1, 15: 3},
                         helpers.added_lines('@@+1\n+wah\n@@+15\n+foo'))

    def test_added_lines_works_with_github_no_newline_message(self):
        patch = """@@ -0,0 +1,5 @@
+class SomeClass
+  def yo(some_unused_param)
+    puts 'hi'
+  end
+end
\ No newline at end of file"""
        try:
            helpers.added_lines(patch)
        except AssertionError:
            self.fail('added_lines() raised AssertionError')


class PluralTest(unittest.TestCase):
    def test_plural__with_one__int(self):
        self.assertEqual('1 unit', helpers.plural(1, 'unit'))

    def test_plural__with_one__list(self):
        self.assertEqual('1 unit', helpers.plural([1], 'unit'))

    def test_plural__with_two__int(self):
        self.assertEqual('2 units', helpers.plural(2, 'unit'))

    def test_plural__with_two__list(self):
        self.assertEqual('2 units', helpers.plural([1, 2], 'unit'))

    def test_plural__with_zero__int(self):
        self.assertEqual('0 units', helpers.plural(0, 'unit'))

    def test_plural__with_zero__list(self):
        self.assertEqual('0 units', helpers.plural([], 'unit'))


class ParseBool(unittest.TestCase):
    def test_parse_bool__non_string(self):
        for value in [True, -1, 1, 1000, [''], 1.0]:
            self.assertTrue(helpers.parse_bool(value))
        for value in [None, False, 0, '', [], {}, 0.0]:
            self.assertFalse(helpers.parse_bool(value))

    def test_parse_bool__string(self):
        for value in '1 on ON On oN t true y yes'.split():
            self.assertTrue(helpers.parse_bool(value))
        for value in '0 off OFF f false n no other'.split():
            self.assertFalse(helpers.parse_bool(value))


class ParseSet(unittest.TestCase):
    def test_parse_set__comma_separated_as_string(self):
        self.assertEqual({'bar', 'bAz', 'foo'},
                         helpers.parse_set('foo, bar ,bAz'))

    def test_parse_set__comma_separated_in_list(self):
        self.assertEqual({'bar', 'baz', 'foo'},
                         helpers.parse_set(['foo, bar ,baz']))

    def test_parse_set__normalize(self):
        self.assertEqual({'hello'}, helpers.parse_set('HELLO', normalize=True))

    def test_parse_set__empty_input(self):
        self.assertEqual(None, helpers.parse_set([]))
        self.assertEqual(None, helpers.parse_set([',', '']))
        self.assertEqual(None, helpers.parse_set(''))
        self.assertEqual(None, helpers.parse_set(' '))

    def test_parse_set__separate_items(self):
        self.assertEqual(set(['bar', 'foo']),
                         helpers.parse_set(['foo', 'bar']))


class PromptTest(unittest.TestCase):
    @patch('farcy.helpers.sys.stdin')
    @patch('farcy.helpers.sys.stdout')
    def test_prompt(self, mock_stdout, mock_stdin):
        mock_stdin.readline.return_value = ' hello '
        self.assertEqual('hello', helpers.prompt('my message'))
        mock_stdout.write.assert_called_with('my message: ')
        self.assertTrue(mock_stdout.flush.called)
